@@ -5,13 +5,13 @@
  * By James D. Berry <jberry@macports.org> 4/6/06
  *
  */
- 
+
 #if HAVE_CONFIG_H
 #include <config.h>
 #endif
 
 /* required for seteuid(2)/setegid(2) */
-#define _BSD_SOURCE
+#define _DEFAULT_SOURCE
 
 #include <sys/types.h>
 #include <grp.h>
@@ -28,7 +28,7 @@
 
 /*
 	getuid
-	
+
 	synopsis: getuid
 */
 int getuidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
@@ -38,14 +38,14 @@ int getuidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Ob
         Tcl_WrongNumArgs(interp, 1, objv, NULL);
 		return TCL_ERROR;
 	}
-	
+
 	Tcl_SetObjResult(interp, Tcl_NewLongObj(getuid()));
 	return TCL_OK;
 }
 
 /*
 	geteuid
-	
+
 	synopsis: geteuid
 */
 int geteuidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
@@ -55,7 +55,7 @@ int geteuidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_O
         Tcl_WrongNumArgs(interp, 1, objv, NULL);
 		return TCL_ERROR;
 	}
-	
+
 	Tcl_SetObjResult(interp, Tcl_NewLongObj(geteuid()));
 	return TCL_OK;
 }
@@ -69,7 +69,7 @@ int getgidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Ob
         Tcl_WrongNumArgs(interp, 1, objv, NULL);
         return TCL_ERROR;
     }
-    
+
     Tcl_SetObjResult(interp, Tcl_NewLongObj(getgid()));
     return TCL_OK;
 }
@@ -83,68 +83,107 @@ int getegidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_O
         Tcl_WrongNumArgs(interp, 1, objv, NULL);
         return TCL_ERROR;
     }
-    
+
     Tcl_SetObjResult(interp, Tcl_NewLongObj(getegid()));
     return TCL_OK;
 }
 
+static int initGroupsForUID(Tcl_Interp *interp, uid_t uid)
+{
+    struct passwd *pwent = getpwuid(uid);
+    if (!pwent) {
+        char buffer[128];
+        snprintf(buffer, sizeof(buffer), "getpwuid failed for uid %d: %d %s", uid, errno, strerror(errno));
+        Tcl_SetObjResult(interp, Tcl_NewStringObj(buffer, -1));
+        return TCL_ERROR;
+    }
+
+    if (initgroups(pwent->pw_name, pwent->pw_gid)) {
+        char buffer[128];
+        snprintf(buffer, sizeof(buffer), "initgroups(%s, %d) failed: %d %s", pwent->pw_name, pwent->pw_gid, errno, strerror(errno));
+        Tcl_SetObjResult(interp, Tcl_NewStringObj(buffer, -1));
+        return TCL_ERROR;
+    }
+    return TCL_OK;
+}
+
 /*
 	setuid
-	
+
 	synopsis: setuid uid
 */
 int setuidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-	long uid = 0;
-	
 	/* Check the arg count */
 	if (objc != 2) {
 		Tcl_WrongNumArgs(interp, 1, objv, "uid");
 		return TCL_ERROR;
 	}
-	
+
 	/* Get the new uid */
-	if (TCL_OK != Tcl_GetLongFromObj(interp, objv[1], &uid))
+	long tmpuid; /* Tcl has no unsigned versions of its get functions. */
+	if (TCL_OK != Tcl_GetLongFromObj(interp, objv[1], &tmpuid)) {
 		return TCL_ERROR;
-		
+	}
+	uid_t uid = (uid_t)tmpuid;
+
+    /* Set up group membership corresponding to the new uid -
+       must be done as root. */
+    if (initGroupsForUID(interp, uid) != TCL_OK) {
+        return TCL_ERROR;
+    }
+
 	/* set the uid */
 	if (0 != setuid(uid)) {
         char buffer[128];
-        snprintf(buffer, sizeof(buffer), "could not set uid to %ld: %d %s", uid, errno, strerror(errno));
+        snprintf(buffer, sizeof(buffer), "could not set uid to %d: %d %s", uid, errno, strerror(errno));
         Tcl_SetObjResult(interp, Tcl_NewStringObj(buffer, -1));
         return TCL_ERROR;
     }
-		
+
 	return TCL_OK;
 }
 
 /*
 	seteuid
-	
+
 	synopsis: seteuid uid
 */
 int seteuidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-	long uid = 0;
-
 	/* Check the arg count */
 	if (objc != 2) {
 		Tcl_WrongNumArgs(interp, 1, objv, "uid");
 		return TCL_ERROR;
 	}
-	
+
 	/* Get the new euid */
-	if (TCL_OK != Tcl_GetLongFromObj(interp, objv[1], &uid))
+	long tmpuid; /* Tcl has no unsigned versions of its get functions. */
+	if (TCL_OK != Tcl_GetLongFromObj(interp, objv[1], &tmpuid)) {
 		return TCL_ERROR;
-		
+	}
+	uid_t uid = (uid_t)tmpuid;
+
+    /* Set up group membership corresponding to the new uid -
+       must be done as root. */
+    uid_t old_euid = geteuid();
+    if (old_euid == 0 && initGroupsForUID(interp, uid) != TCL_OK) {
+        return TCL_ERROR;
+    }
+
 	/* set the euid */
 	if (0 != seteuid(uid)) {
         char buffer[128];
-        snprintf(buffer, sizeof(buffer), "could not set effective uid to %ld: %d %s", uid, errno, strerror(errno));
+        snprintf(buffer, sizeof(buffer), "could not set effective uid to %d: %d %s", uid, errno, strerror(errno));
         Tcl_SetObjResult(interp, Tcl_NewStringObj(buffer, -1));
         return TCL_ERROR;
     }
-		
+
+    /* Run initgroups if we are root now but weren't before. */
+    if (uid == 0 && old_euid != 0 && initGroupsForUID(interp, uid) != TCL_OK) {
+        return TCL_ERROR;
+    }
+
 	return TCL_OK;
 }
 
@@ -154,23 +193,23 @@ int seteuidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_O
 int setgidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
     long gid;
-    
+
     if (objc != 2) {
         Tcl_WrongNumArgs(interp, 1, objv, "gid");
         return TCL_ERROR;
     }
-    
+
     if (TCL_OK != Tcl_GetLongFromObj(interp, objv[1], &gid)) {
         return TCL_ERROR;
     }
-    
+
     if (0 != setgid(gid)) {
         char buffer[128];
         snprintf(buffer, sizeof(buffer), "could not set gid to %ld: %d %s", gid, errno, strerror(errno));
         Tcl_SetObjResult(interp, Tcl_NewStringObj(buffer, -1));
         return TCL_ERROR;
     }
-    
+
     return TCL_OK;
 }
 
@@ -180,23 +219,23 @@ int setgidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Ob
 int setegidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
     long gid;
-    
+
     if (objc != 2) {
         Tcl_WrongNumArgs(interp, 1, objv, "gid");
         return TCL_ERROR;
     }
-    
+
     if (TCL_OK != Tcl_GetLongFromObj(interp, objv[1], &gid)) {
         return TCL_ERROR;
     }
-    
+
     if (0 != setegid(gid)) {
         char buffer[128];
         snprintf(buffer, sizeof(buffer), "could not set effective gid to %ld: %d %s", gid, errno, strerror(errno));
         Tcl_SetObjResult(interp, Tcl_NewStringObj(buffer, -1));
         return TCL_ERROR;
     }
-    
+
     return TCL_OK;
 }
 
@@ -307,56 +346,56 @@ int getpwuidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_
 
 /*
 	name_to_uid
-	
+
 	synopsis: name_to_uid name
 */
 int name_to_uidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
 	struct passwd *pwent;
 	char* name = NULL;
-	
+
 	/* Check the arg count */
 	if (objc != 2) {
 		Tcl_WrongNumArgs(interp, 1, objv, "name");
 		return TCL_ERROR;
 	}
-	
+
 	/* Get the  name */
 	name = Tcl_GetString(objv[1]);
 	if (name == NULL || !*name)
 		return TCL_ERROR;
-	
+
 	/* Map the name --> uid */
 	pwent = getpwnam(name);
 
 	if (pwent == NULL)
 		Tcl_SetObjResult(interp, Tcl_NewIntObj(-1));
 	else
-		Tcl_SetObjResult(interp, Tcl_NewIntObj(pwent->pw_uid)); 
+		Tcl_SetObjResult(interp, Tcl_NewIntObj(pwent->pw_uid));
 
 	return TCL_OK;
 }
 
 /*
 	uid_to_name
-	
+
 	synopsis: uid_to_name uid
 */
 int uid_to_nameCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
 	long uid = 0;
 	struct passwd *pwent;
-	
+
 	/* Check the arg count */
 	if (objc != 2) {
 		Tcl_WrongNumArgs(interp, 1, objv, "uid");
 		return TCL_ERROR;
 	}
-	
+
 	/* Get the  uid */
 	if (TCL_OK != Tcl_GetLongFromObj(interp, objv[1], &uid))
 		return TCL_ERROR;
-	
+
 	/* Map the uid --> name, or empty result on error */
 	pwent = getpwuid(uid);
 	if (pwent != NULL)
@@ -367,7 +406,7 @@ int uid_to_nameCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, T
 
 /*
 	uname_to_gid
-	
+
 	synopsis: uname_to_gid name
 	this function takes a *user* name
 */
@@ -375,25 +414,25 @@ int uname_to_gidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc,
 {
 	struct passwd *pwent;
 	char* name = NULL;
-	
+
 	/* Check the arg count */
 	if (objc != 2) {
 		Tcl_WrongNumArgs(interp, 1, objv, "name");
 		return TCL_ERROR;
 	}
-	
+
 	/* Get the  name */
 	name = Tcl_GetString(objv[1]);
 	if (name == NULL || !*name)
 		return TCL_ERROR;
-	
+
 	/* Map the name --> user gid */
 	pwent = getpwnam(name);
 
 	if (pwent == NULL)
 		Tcl_SetObjResult(interp, Tcl_NewIntObj(-1));
 	else
-		Tcl_SetObjResult(interp, Tcl_NewIntObj(pwent->pw_gid)); 
+		Tcl_SetObjResult(interp, Tcl_NewIntObj(pwent->pw_gid));
 
 	return TCL_OK;
 }
@@ -408,23 +447,23 @@ int name_to_gidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, T
 {
     struct group *grent;
     char *name;
-    
+
     if (objc != 2) {
         Tcl_WrongNumArgs(interp, 1, objv, "name");
         return TCL_ERROR;
     }
-    
+
     name = Tcl_GetString(objv[1]);
     if (name == NULL || !*name)
         return TCL_ERROR;
-    
+
     grent = getgrnam(name);
-    
+
     if (grent == NULL)
         Tcl_SetObjResult(interp, Tcl_NewIntObj(-1));
     else
-        Tcl_SetObjResult(interp, Tcl_NewIntObj(grent->gr_gid)); 
-    
+        Tcl_SetObjResult(interp, Tcl_NewIntObj(grent->gr_gid));
+
     return TCL_OK;
 }
 
@@ -435,18 +474,18 @@ int gid_to_nameCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, T
 {
     long gid;
     struct group *grent;
-    
+
     if (objc != 2) {
         Tcl_WrongNumArgs(interp, 1, objv, "gid");
         return TCL_ERROR;
     }
-    
+
     if (TCL_OK != Tcl_GetLongFromObj(interp, objv[1], &gid))
         return TCL_ERROR;
-    
+
     grent = getgrgid(gid);
     if (grent != NULL)
         Tcl_SetResult(interp, grent->gr_name, TCL_STATIC);
-    
+
     return TCL_OK;
 }
